* SingularValues.F
* singular value decomposition of an m-by-n matrix
* last modified 11 Jul 06 th

#include "diag.h"


************************************************************************
** SingularValues performs a singular value decomposition.
** Input: m, n, A = m-by-n matrix (m > n).
** Output: d = n-vector of singular values,
** V = n-by-m left transformation matrix,
** W = n-by-n right transformation matrix,
** these fulfill diag(d) = V^* A W^+.

	subroutine SingularValues(m, n, A,ldA, d, V,ldV, W,ldW, sort)
	implicit none
	integer m, n, ldA, ldV, ldW, sort
	double complex A(ldA,n), V(ldV,m), W(ldW,n)
	double precision d(n)

	integer i, j, k, l, nz, iz(MAXMAT)
	logical valid(MAXMAT)
	double complex s
	double precision norm

	double complex AdagA(MAXMAT,MAXMAT)

	double precision eps
	parameter (eps = 5D-16)

* The singular values are the roots of the eigenvalues of A^+ A

	do k = 1, n
	  do i = 1, k
	    s = 0
	    do j = 1, m
	      s = s + DCONJG(A(j,i))*A(j,k)
	    enddo
	    AdagA(i,k) = s
	  enddo
	enddo
	call Eigensystem(n, AdagA,MAXMAT, d, W,ldW, sort)

	nz = 0
	do k = 1, n
	  valid(k) = d(k) .gt. eps
	  if( valid(k) ) then
	    d(k) = sqrt(d(k))
	    norm = 1/d(k)
	    do i = 1, m
	      s = 0
	      do j = 1, n
	        s = s + A(i,j)*DCONJG(W(k,j))
	      enddo
	      V(k,i) = s*norm
	    enddo
	  else
	    nz = nz + 1
	    iz(nz) = k
	  endif
	enddo

* The vectors corresponding to the zero singular values can be chosen
* arbitrarily.  We choose u_i = delta_ij and use Gram-Schmidt to
* orthonormalize these.

#define Delta(i,j) ibits(ieor(i,j)-1,15,1)

	l = m
	do i = 1, nz
	  norm = 0

1	  do j = 1, m
	    s = 0
	    do k = 1, n
	      if( valid(k) ) s = s + V(k,l)*DCONJG(V(k,j))
	    enddo
	    s = Delta(l, j) - s
	    V(iz(i),j) = s
	    norm = norm + DBLE(s)**2 + DIMAG(s)**2
	  enddo
	  valid(iz(i)) = .TRUE.
	  l = l - 1
	  if( norm .lt. eps ) goto 1

	  norm = 1/sqrt(norm)
	  k = iz(i)
	  do j = 1, m
	    V(k,j) = norm*V(k,j)
	  enddo
	enddo
	end

